/*
   Siag, Scheme In A Grid
   Copyright (C) 1996-2000  Ulric Eriksson <ulric@siag.nu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <unistd.h>
#include <sys/stat.h>

#include <X11/Intrinsic.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>

#include "../siod/siod.h"

#include "../common/cmalloc.h"
#include "../common/common.h"
#include "../common/fonts.h"

#include "../xcommon/xcommon.h"

#include "../siag/calc.h"

#include "gsiag.h"
#include "sgrid.h"

window *w_list;
int pr_scr_flag;
int grid_lines = 1;

int lastc, macro_flag;

GtkWidget *topLevel;

static GtkWidget *topbox, *menu_bar, *toolbar, *toolbarhandlebox,
    *formatbar, *formatbarhandlebox, *gridpane, *text1, *label1, *label2;

static GtkWidget *btnFont, *mnuFont, *btnSize, *mnuSize, *btnStyle,
    *mnuStyle, *btnColor, *mnuColor;
static GtkWidget *cmdBold, *cmdItalic, *cmdHLeft, *cmdHCenter, *cmdHRight;
static GtkWidget *cmdBorders, *cmdGrid,
	*cmdUline, *cmdLline, *cmdRline, *cmdNone;

static GtkItemFactory *menu_factory = NULL;

#if 0
	factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", grp);
	gtk_item_factory_create_items_ac(factory, nmenu_items, menu_items, NULL, 2);

Men det r inte s vi vill gra, utan inkrementellt bygga upp menystrukturen
genom Scheme-anrop. Det finns ett antal olika stt en GtkItemFactoryEntry
kan se ut:

{ "/_File", NULL, 0, 0, "<Branch>" }
{ "/File/tearoff1", NULL, callback, 0 }
{ "/File/New", "<control>N", callback, 0 }
{ "/File/sep1", NULL, callback, 0, "<Separator>" }
{ "/File/Submenu", NULL, 0, 0, "<Branch>" }
{ "/Help", NULL, 0, 0, "<LastBranch>" }
{ "/Help/About", NULL, callback, 0 }

Hr fattas callbackdata som r ndvndigt fr att vr menukod skall
fungera. Jag har en stark knsla av att nullan efter callback mste
ndras till en etta fr att man skall kunna skicka med data.

Hr r exempelkod inspirerad av testgtk.c:

	accel_group = gtk_accel_group_new();
	item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", accel_group);
	gtk_accel_group_attach(accel_group, GTK_OBJECT(window));
	gtk_item_factory_create_items(item_factory, nmenu_items,
			menu_items, NULL);
	gtk_box_pack_start(GTK_BOX(box1),
		gtk_item_factory_get_widget(item_factory, "<main>"),
		FALSE, FALSE, 0);

Men vi mste anvnda gtk_item_factory_create_item:

	gtk_item_factory_create_item(item_factory, entry,
		 callback_data, callback_type);

Av ngon anledning r det bra om callback_type r 1. Vet ej vad
det betyder.

Jag struntar nu i understrukna tecken och acceleratorer. Jag
struntar ocks i hgerklickmenyn.
#endif

float zoom = 1.0;

static GtkTooltips *helptext;

extern int recalc;
int input_warp_pointer = 1;	/* move pointer to input field */

static int status;

static void draw_input(char *text)
{
	int r, c;
	int s;
	int type, intp;
	char b[1024];

	r = get_point(w_list).row;
	c = get_point(w_list).col;
	s = w_list->sht;
        sprintf(b, "[%d,%d]", r, c);
	gtk_label_set(GTK_LABEL(label1), b);
        type = ret_type(w_list->buf, s, r, c);
        intp = ret_interpreter(w_list->buf, s, r, c);
        switch (type) {
        case EXPRESSION:
        case STRING:
                sprintf(b, "%s", interpreter2name(intp));
                break;
        case LABEL:
                strcpy(b, "LABEL");
                break;
        case CONSTANT:
                strcpy(b, "CONSTANT");
                break;
        case EMPTY:
                strcpy(b, "EMPTY");
                break;
        default:
                strcpy(b, "ERROR");
        }
}

static void draw_status(char *text)
{
	char *p = strchr(text, '\n');
	if (p)
		*p = '\0';
	gtk_label_set(GTK_LABEL(label2), text);
}

void llpr(char *p)
{
	static int isclear = FALSE;

	if (isclear && p[0] == '\0')
		return;
	isclear = (p[0] == '\0');

	draw_status(p);
}

static gint cell_format(GtkSGrid * dummy, gpointer data,
			gint row, gint col)
{
	window *w = (window *)data;
	return ret_format(w->buf, w->sht, row, col);
}

static gint ng_col_width(GtkSGrid * dummy, gpointer data, gint col)
{
	return 80;
}

static gint ng_row_height(GtkSGrid * dummy, gpointer data, gint row)
{
	return 20;
}

static gchar *ng_text(GtkSGrid * dummy, gpointer b, gint row, gint col)
{
	window *w = (window *)b;
	static char s[1024];
	s[0] = '\0';
	ret_pvalue(s, w->buf, w->sht, row, col, -1);
	return s;
}

static void draw_cells(window * w)
{
	int top_row, top_col;
	position prot, top, blku, blkl, point;
	prot = get_prot(w);
	top = get_top(w);
	if (w->sht == w->bsht) {
		blku = block_upper(w);
		blkl = block_lower(w);
	} else {
		blku.row = blku.col = blkl.row = blkl.col = -1;
	}
	point = get_point(w);
	GTK_SGRID(w->ui->grid)->max_row = BUFFER_ROWS;
	GTK_SGRID(w->ui->grid)->max_col = BUFFER_COLS;
	GTK_SGRID(w->ui->grid)->prot_row = prot.row;
	GTK_SGRID(w->ui->grid)->prot_col = prot.col;
	GTK_SGRID(w->ui->grid)->top_row = top.row;
	GTK_SGRID(w->ui->grid)->top_col = top.col;
	GTK_SGRID(w->ui->grid)->sel_top_row = blku.row;
	GTK_SGRID(w->ui->grid)->sel_bottom_row = blkl.row;
	GTK_SGRID(w->ui->grid)->sel_left_col = blku.col;
	GTK_SGRID(w->ui->grid)->sel_right_col = blkl.col;
	GTK_SGRID(w->ui->grid)->point_row = point.row;
	GTK_SGRID(w->ui->grid)->point_col = point.col;
	GTK_SGRID(w->ui->grid)->row_height = ng_row_height;
	GTK_SGRID(w->ui->grid)->col_width = ng_col_width;
/*	GTK_SGRID(w->ui->grid)->type =*/
	GTK_SGRID(w->ui->grid)->text = ng_text;
/*	GTK_SGRID(w->ui->grid)->format =*/
	GTK_SGRID(w->ui->grid)->data = w;
	GTK_SGRID(w->ui->grid)->grid_lines = grid_lines;
	gtk_sgrid_set_point(GTK_SGRID(w->ui->grid),
			    point.row, point.col);

	/* there is a chance that this moved top */
	top_row = GTK_SGRID(w->ui->grid)->top_row;
	top_col = GTK_SGRID(w->ui->grid)->top_col;
	if (top_row != top.row) {
		set_top_row(w, top_row);
	}
	if (top_col != top.col) {
		set_top_col(w, top_col);
	}
}

static int dialog_key(GtkWidget * w, GdkEventKey * event,
		      gpointer user_data)
{
	int c = event->keyval;
	if (c == GDK_Escape)
		status = ABORT;
	if (c == GDK_Return)
		status = DONE;
	return TRUE;
}

int ask_for_str_comp(char *prompt, char *buffr, int (*comp) (char *))
{
	char *p;
	status = WAITING;

	gtk_label_set(GTK_LABEL(label1), prompt);
	gtk_entry_set_text(GTK_ENTRY(text1), buffr);
	gtk_widget_grab_focus(text1);
	gtk_grab_add(text1);
	while (status == WAITING)
		gtk_main_iteration();

	p = gtk_entry_get_text(GTK_ENTRY(text1));
	if (p)
		strcpy(buffr, p);
	gtk_entry_set_text(GTK_ENTRY(text1), "");
	gtk_grab_remove(text1);
	activate_window(w_list);
	return status;
}

int edit_cell(char *p, char *b)
{
	return ask_for_str(p, b);
}

int add_str_to_input_queue(textbuf buf)
{
	return FALSE;
}

static void set_block(int row1, int col1, int row2, int col2)
{
        int ur = block_upper(w_list).row, uc = block_upper(w_list).col;
        int lr = block_lower(w_list).row, lc = block_lower(w_list).col;
 
        w_list->bsht = w_list->sht;
 
        if (row1 < 1) row1 = 1;
        if (row1 > BUFFER_ROWS) row1 = BUFFER_ROWS;
        if (col1 < 1) col1 = 1;
        if (col1 > BUFFER_COLS) col1 = BUFFER_COLS;
        if (row2 < 1) row2 = 1;
        if (row2 > BUFFER_ROWS) row2 = BUFFER_ROWS;
        if (col2 < 1) col2 = 1;
        if (col2 > BUFFER_COLS) col2 = BUFFER_COLS;
 
        if (row1 < row2) {
                set_blku_row(w_list, row1);
                set_blkl_row(w_list, row2);
        } else {
                set_blku_row(w_list, row2);
                set_blkl_row(w_list, row1);
        }
        if (col1 < col2) {
                set_blku_col(w_list, col1);
                set_blkl_col(w_list, col2);
        } else {
                set_blku_col(w_list, col2);
                set_blkl_col(w_list, col1);
        }
 
        /* Redraw if any change */
        if (ur != block_upper(w_list).row || uc != block_upper(w_list).col ||
                lr != block_lower(w_list).row || lc != block_lower(w_list).col) {
		pr_scr_flag = TRUE;
        }
}

#ifndef ABS
#define ABS(a) ((a)>0?(a):-(a))
#endif
#if 0
static LISP lhelptext_mode(LISP newmode)
{
	return NIL;
}
#endif

static LISP ltooltip_mode(LISP newmode)
{
        return NIL;
}

static void execute_callback(GtkWidget * w, char *client_data)
{
	execute((char *) client_data);
}

static void factory_callback(gpointer callback_data,
		guint callback_action, GtkWidget *widget)
{
	char *p = (char *)callback_data;
	execute(p);
}

static LISP add_menu(LISP lisp_label)
{
	GtkItemFactoryEntry entry = { NULL, NULL, 0, 0, "<Branch>" };
	char b[1024];
	char *lbl = _(get_c_string(lisp_label));

	sprintf(b, "/%s", lbl);
	entry.path = b;
	if (!strcmp(lbl, _("Help"))) entry.item_type = "<LastBranch>";

	gtk_item_factory_create_item(menu_factory, &entry, NULL, 1);

	return NIL;
}

static LISP add_menu_entry(LISP menu, LISP label, LISP function)
{
	GtkItemFactoryEntry entry = { NULL, NULL, factory_callback, 0, NULL };
	char b[1024];
	char *mnu = _(get_c_string(menu)), *lbl = _(get_c_string(label));
	char *fun = get_c_string(function);

	sprintf(b, "/%s/%s", mnu, lbl);
	entry.path = b;

	if (!strcmp(lbl, "-")) {	/* line pane */
		entry.item_type = "<Separator>";
	}
	gtk_item_factory_create_item(menu_factory, &entry, fun, 1);
	return NIL;
}

static LISP add_submenu(LISP menu, LISP label)
{
	GtkItemFactoryEntry entry = { NULL, NULL, 0, 0, "<Branch>" };
	char b[1024];
	char *mnu = _(get_c_string(menu));
	char *lbl = _(get_c_string(label));

	sprintf(b, "/%s/%s", mnu, lbl);
	entry.path = b;

	gtk_item_factory_create_item(menu_factory, &entry, NULL, 1);

	return NIL;
}

static LISP add_submenu_entry(LISP menu, LISP submenu, LISP label,
			      LISP function)
{
	GtkItemFactoryEntry entry = { NULL, NULL, factory_callback, 0, NULL };
	char b[1000];
	char *mnu = _(get_c_string(menu));
	char *sm = _(get_c_string(submenu));
	char *lbl = _(get_c_string(label));
	char *fun = get_c_string(function);

	sprintf(b, "/%s/%s/%s", mnu, sm, lbl);
	entry.path = b;
	if (!strcmp(lbl, "-"))
		entry.item_type = "<Separator>";
	gtk_item_factory_create_item(menu_factory, &entry, fun, 1);
	return NIL;
}

static void find_pixmap(char *fn, char *pm)
{
	char *pixpath = getenv("PIXPATH");
        struct stat statbuf;
        int found = 0;
	if (pm[0] == '/') {
		strcpy(fn, pm);
		found = 1;
	} else if (pixpath) {
		sprintf(fn, "%s/%s", pixpath, pm);
		found = !stat(fn, &statbuf);
	}
	if (!found) sprintf(fn, "%s/common/bitmaps/kde/%s", datadir, pm);
	found = !stat(fn, &statbuf);
        if (!found) sprintf(fn, "%s/common/bitmaps/%s", datadir, pm);
}

static GtkWidget *make_toggle(char *cmd, GtkWidget *pw, char *pm, char *t)
{
	GtkWidget *w = gtk_toggle_button_new(), *pmw;
	GdkWindow *win = pw->window;
	GdkBitmap *mask = NULL;
	GdkPixmap *gpm;
	GtkStyle *style = gtk_widget_get_default_style();
	char fn[1024];
	find_pixmap(fn, pm);
	gpm = gdk_pixmap_create_from_xpm(win,
					 &mask,
					 &style->bg[GTK_STATE_NORMAL],
					 fn);
	pmw = gtk_pixmap_new(gpm, mask);

	gtk_container_add(GTK_CONTAINER(w), pmw);
	gtk_widget_show(pmw);
	/* gtk_box_pack_start(GTK_BOX(pw), w, FALSE, FALSE, 2); */
	gtk_widget_show(w);
	gtk_signal_connect_object(GTK_OBJECT(w),
				  "clicked",
				  GTK_SIGNAL_FUNC(execute_callback),
				  (gpointer) g_strdup(cmd));

	gtk_toolbar_append_widget(GTK_TOOLBAR(formatbar),
				  GTK_WIDGET(w), t, NULL);

	GTK_WIDGET_UNSET_FLAGS(w, GTK_CAN_FOCUS);
	GTK_WIDGET_UNSET_FLAGS(w, GTK_CAN_DEFAULT);
	return w;
}

/*changed RW */

static GtkWidget *make_command(char *cmd, GtkWidget *pw, char *pm, char *t)
{
	/*GtkWidget *w = gtk_button_new(); */
	GtkWidget *pmw, *w;
	GdkWindow *win = pw->window;
	GdkBitmap *mask = NULL;
	GtkStyle *style = gtk_widget_get_default_style();
	GdkPixmap *gpm;
	char fn[1024];
	find_pixmap(fn, pm);
	gpm = gdk_pixmap_create_from_xpm(win,
					 &mask,
					 &style->bg[GTK_STATE_NORMAL],
					 fn);
	pmw = gtk_pixmap_new(gpm, mask);
	w = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
				    NULL,
				    t,
				    NULL,
				    GTK_WIDGET(pmw),
				    GTK_SIGNAL_FUNC(execute_callback),
				    (gpointer) g_strdup(cmd));

	return w;
}
/* stopped changing */

static GtkWidget *make_vsep(GtkWidget *pw)
{
	return NULL;
}

static void init_toolbar(void)
{
#if 1
        make_command("(new-buffer)", toolbar, "new.xpm",
                "Start another instance of Siag");
        make_command("(load-buffer)", toolbar, "fld_open.xpm",
                "Open a Siag document");
        make_command("(save-buffer-as)", toolbar, "save.xpm",
                "Save the contents of the current buffer");
        make_command("(preview)", toolbar, "preview.xpm",
                "Preview the contents of the current buffer");
        make_command("(print)", toolbar, "printer.xpm",
                "Print the contents of the current buffer");
        make_vsep(toolbar);
        make_command("(cut-block-to-string)", toolbar, "cut.xpm",
                "Cut");
        make_command("(copy-block-to-string)", toolbar, "copy.xpm",
                "Copy");
        make_command("(paste-block-from-string)", toolbar, "paste.xpm",
                "Paste");
        make_command("(undo-restore)", toolbar, "undo.xpm",
                "Undo");
        make_command("(sort-block 0 nil nil nil)", toolbar, "sortaz.xpm",
                "Sort ascending");
        make_command("(sort-block 0 nil t nil)", toolbar, "sortza.xpm",
                "Sort descending");
        make_command("(plot-wizard)", toolbar, "plotter.xpm",
                "Plot the contents of the block using lines");
        make_command("(block-sum)", toolbar, "sigma.xpm",
                "Add the contents of the block");
        make_vsep(toolbar);
        make_command("(help-contents)", toolbar, "info.xpm",
                "Display the Siag online documentation");
        make_command("(help-copyright)", toolbar, "copyright.xpm",
                "Display the Gnu general public license");
#else
	    make_command("tbOpen", "(load-buffer)", toolbar,
			 (char **) fld_open_xpm, "Open a Siag document");

	    make_command("tbSave", "(save-buffer-as)", toolbar,
			 (char **) floppy3_xpm,
			 "Save the contents of the current buffer");

	    make_command("tbView", "(preview)", toolbar,
			 (char **) preview_xpm,
			 "Preview the contents of the current buffer");

	    make_command("tbPrint", "(print)", toolbar,
			 (char **) printer_xpm,
			 "Print the contents of the current buffer");

	    make_command("tbPlot", "(plot \"lines\")", toolbar,
			 (char **) plotter_xpm,
			 "Plot the contents of the block using lines");

	    make_command("tbSum", "(block-sum)", toolbar,
			 (char **) sigma_xpm,
			 "Add the contents of the block");

	    make_command("tbHelp", "(help-contents)", toolbar,
			 (char **) info_xpm,
			 "Display the Siag online documentation");

	    make_command("tbCopy", "(help-copyright)", toolbar,
			 (char **) copyright_xpm,
			 "Display the GNU general public licence");

	    make_command("cmdBorders", "(block-borders 1)", toolbar,
			 (char **) borders_xpm,
			 "Draw borders around the block");

	    make_command("cmdGrid", "(block-borders 2)", toolbar,
			 (char **) grid_xpm,
			 "Draw grid lines in the block");

	    make_command("cmdUline", "(block-borders 3)", toolbar,
			 (char **) uline_xpm, "Underline the block");

	    make_command("cmdNone", "(block-borders 0)", toolbar,
			 (char **) none_xpm,
			 "Remove grid lines from the block");
#endif
}

static void init_toggle()
{
        cmdBold = make_toggle("(toggle-format \"bold\")",
                formatbar, "bold.xpm", "Bold text");
        cmdItalic = make_toggle("(toggle-format \"italic\")",
                formatbar, "italic.xpm", "Italic text");
        make_vsep(formatbar);
        cmdHLeft = make_toggle("(new-format \"hadj\" HADJ_LEFT)",
                formatbar, "hleft.xpm", "Left adjusted text");
        cmdHCenter = make_toggle("(new-format \"hadj\" HADJ_CENTER)",
                formatbar, "hcenter.xpm", "Centered text");
        cmdHRight = make_toggle("(new-format \"hadj\" HADJ_RIGHT)",
                formatbar, "hright.xpm", "Right adjusted text");
        make_vsep(toolbar);
        cmdBorders = make_command("(block-borders 1)", toolbar,
                "borders.xpm", "Draw borders around the block");
        cmdGrid = make_command("(block-borders 2)", toolbar,
                "grid.xpm", "Draw grid lines in the block");
        cmdUline = make_command("(block-borders 3)", toolbar,
                "uline.xpm", "Underline the block");
        cmdLline = make_command("(block-borders 5)", toolbar,
                "lline.xpm", "Leftline the block");
        cmdRline = make_command("(block-borders 6)", toolbar,
                "rline.xpm", "Rightline the block");
        cmdNone = make_command("(block-borders 0)", toolbar,
                "none.xpm", "Remove grid lines from the block");
}

static void font_menu(GtkWidget * w, char *label, char *cmd)
{
	GtkWidget *entry = gtk_menu_item_new_with_label(label);
	gtk_menu_append(GTK_MENU(w), entry);
	gtk_signal_connect_object(GTK_OBJECT(entry),
				  "activate",
				  GTK_SIGNAL_FUNC(execute_callback),
				  (gpointer) g_strdup(cmd));
	gtk_widget_show(entry);
}

static GtkWidget *create_font_menu()
{
	mnuFont = gtk_menu_new();
	font_menu(mnuFont, "Courier", "(change-font COURIER FONT_MASK)");
	font_menu(mnuFont, "Helvetica",
		  "(change-font HELVETICA FONT_MASK)");
	font_menu(mnuFont, "New Century Schoolbook",
		  "(change-font NEW_CENTURY FONT_MASK)");
	font_menu(mnuFont, "Times", "(change-font TIMES FONT_MASK)");
	return mnuFont;
}

static GtkWidget *create_size_menu()
{
	mnuSize = gtk_menu_new();
	font_menu(mnuSize, "8", "(change-font SIZE_8 SIZE_MASK)");
	font_menu(mnuSize, "10", "(change-font SIZE_10 SIZE_MASK)");
	font_menu(mnuSize, "12", "(change-font SIZE_12 SIZE_MASK)");
	font_menu(mnuSize, "14", "(change-font SIZE_14 SIZE_MASK)");
	font_menu(mnuSize, "18", "(change-font SIZE_18 SIZE_MASK)");
	font_menu(mnuSize, "20", "(change-font SIZE_20 SIZE_MASK)");
	font_menu(mnuSize, "24", "(change-font SIZE_24 SIZE_MASK)");
	font_menu(mnuSize, "30", "(change-font SIZE_30 SIZE_MASK)");
	return mnuSize;
}

static GtkWidget *create_style_menu()
{
	mnuStyle = gtk_menu_new();
	font_menu(mnuStyle, "Default",
		  "(change-font FMT_DEFAULT FMT_MASK)");
	font_menu(mnuStyle, "Invisible",
		  "(change-font FMT_INVISIBLE FMT_MASK)");
	font_menu(mnuStyle, "Integer",
		  "(change-font FMT_INTEGER FMT_MASK)");
	font_menu(mnuStyle, "Scientific",
		  "(change-font FMT_SCIENTIFIC FMT_MASK)");
	font_menu(mnuStyle, "Fixed", "(change-font FMT_FIXED FMT_MASK)");
	font_menu(mnuStyle, "Date", "(change-font FMT_DATE FMT_MASK)");
	font_menu(mnuStyle, "Time", "(change-font FMT_TIME FMT_MASK)");
	font_menu(mnuStyle, "Comma", "(change-font FMT_COMMA FMT_MASK)");
	font_menu(mnuStyle, "Percent",
		  "(change-font FMT_PERCENT FMT_MASK)");
	font_menu(mnuStyle, "Hex", "(change-font FMT_HEX FMT_MASK)");
	font_menu(mnuStyle, "Currency",
		  "(change-font FMT_CURRENCY FMT_MASK)");
	font_menu(mnuStyle, "User 1", "(change-font FMT_USER1 FMT_MASK)");
	font_menu(mnuStyle, "User 2", "(change-font FMT_USER2 FMT_MASK)");
	font_menu(mnuStyle, "User 3", "(change-font FMT_USER3 FMT_MASK)");
	font_menu(mnuStyle, "User 4", "(change-font FMT_USER4 FMT_MASK)");
	font_menu(mnuStyle, "User 5", "(change-font FMT_USER5 FMT_MASK)");
	return mnuStyle;
}

static GtkWidget *create_color_menu()
{
	mnuColor = gtk_menu_new();
	font_menu(mnuColor, "Black",
		  "(change-font COLOR_BLACK COLOR_MASK)");
	font_menu(mnuColor, "Red", "(change-font COLOR_RED COLOR_MASK)");
	font_menu(mnuColor, "Green",
		  "(change-font COLOR_GREEN COLOR_MASK)");
	font_menu(mnuColor, "Blue", "(change-font COLOR_BLUE COLOR_MASK)");
	font_menu(mnuColor, "Yellow",
		  "(change-font COLOR_YELLOW COLOR_MASK)");
	font_menu(mnuColor, "Magenta",
		  "(change-font COLOR_MAGENTA COLOR_MASK)");
	font_menu(mnuColor, "Cyan", "(change-font COLOR_CYAN COLOR_MASK)");
	font_menu(mnuColor, "White",
		  "(change-font COLOR_WHITE COLOR_MASK)");
	return mnuColor;
}

static void setup_buttons(void)
{
	btnFont = gtk_option_menu_new();

	gtk_option_menu_set_menu(GTK_OPTION_MENU(btnFont),
				 create_font_menu(btnFont));

	/*  gtk_box_pack_start(GTK_BOX(formatbar), btnFont, FALSE, FALSE, 2);
	   gtk_widget_show(btnFont);
	   gtk_tooltips_set_tip(helptext, btnFont, "Change the font family",NULL);
	 */
	gtk_toolbar_append_widget(GTK_TOOLBAR(formatbar),
				  GTK_WIDGET(btnFont),
				  "Change the font family", NULL);
	gtk_widget_show(btnFont);

	btnSize = gtk_option_menu_new();
	gtk_option_menu_set_menu(GTK_OPTION_MENU(btnSize),
				 create_size_menu(btnSize));

	/* gtk_box_pack_start(GTK_BOX(formatbar), btnSize, FALSE, FALSE, 2);
	   gtk_widget_show(btnSize);
	   gtk_tooltips_set_tip(helptext, btnSize, "Change the font size",NULL);
	 */
	gtk_toolbar_append_widget(GTK_TOOLBAR(formatbar),
				  GTK_WIDGET(btnSize),
				  "Change the font size", NULL);
	gtk_widget_show(btnSize);

	btnStyle = gtk_option_menu_new();
	gtk_option_menu_set_menu(GTK_OPTION_MENU(btnStyle),
				 create_style_menu(btnStyle));
	/*gtk_box_pack_start(GTK_BOX(formatbar), btnStyle, FALSE, FALSE, 2);
	   gtk_widget_show(btnStyle);
	   gtk_tooltips_set_tip(helptext, btnStyle, "Change the display style",NULL);
	 */

	gtk_toolbar_append_widget(GTK_TOOLBAR(formatbar),
				  GTK_WIDGET(btnStyle),
				  "Change the display style", NULL);
	gtk_widget_show(btnStyle);

	btnColor = gtk_option_menu_new();
	gtk_option_menu_set_menu(GTK_OPTION_MENU(btnColor),
				 create_color_menu(btnColor));
	/*
	   gtk_box_pack_start(GTK_BOX(formatbar), btnColor, FALSE, FALSE, 2);
	   gtk_widget_show(btnColor);
	   gtk_tooltips_set_tip(helptext, btnColor, "Change the color",NULL);
	 */

	gtk_toolbar_append_widget(GTK_TOOLBAR(formatbar),
				  GTK_WIDGET(btnColor),
				  "Change the color", NULL);

	gtk_widget_show(btnColor);

}

void mainloop()
{
	gtk_main();
}

static void init_windows1(int *argc, char **argv)
{
	gtk_set_locale();
	gtk_init(argc, &argv);
	gtk_rc_parse("gsiagrc");
	topLevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_usize(topLevel, 700, 420);
	gtk_signal_connect(GTK_OBJECT(topLevel), "destroy",
			   (GtkSignalFunc) execute_callback,
			   (gpointer) "(quit-siag)");
	gtk_signal_connect(GTK_OBJECT(topLevel), "delete_event",
			   (GtkSignalFunc) execute_callback,
			   (gpointer) "(quit-siag)");

	topbox = gtk_vbox_new(FALSE, 0);	/* not homogenous, no spacing */
	gtk_container_add(GTK_CONTAINER(topLevel), topbox);
	gtk_widget_show(topbox);

	menu_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR,
			"<main>", NULL);
	menu_bar = gtk_item_factory_get_widget(menu_factory, "<main>");
	gtk_box_pack_start(GTK_BOX(topbox), menu_bar, FALSE, FALSE, 0);
	gtk_widget_show(menu_bar);

	helptext = gtk_tooltips_new();
	/*changed this bit - RW */
	toolbarhandlebox = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(topbox), toolbarhandlebox, FALSE, FALSE,
			   2);
	gtk_signal_connect(GTK_OBJECT(toolbarhandlebox), "child_attached",
			   NULL, "attached");
	gtk_signal_connect(GTK_OBJECT(toolbarhandlebox), "child_detached",
			   NULL, "detached");
	gtk_widget_show(toolbarhandlebox);

	toolbar =
	    gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
	gtk_container_add(GTK_CONTAINER(toolbarhandlebox), toolbar);
	/*gtk_box_pack_start(GTK_BOX(topbox),toolbar, FALSE, FALSE, 2);
	   toolbar = gtk_hbox_new(FALSE, 0);
	   gtk_container_add (GTK_CONTAINER (realtoolbar), toolbar);
	   gtk_widget_show(toolbar); */
	gtk_widget_show(toolbar);
	init_toolbar();
	/*stopped changing */



	formatbarhandlebox = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(topbox), formatbarhandlebox, FALSE,
			   FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(formatbarhandlebox),
			   "child_attached", NULL, "attached");
	gtk_signal_connect(GTK_OBJECT(formatbarhandlebox),
			   "child_detached", NULL, "detached");
	gtk_widget_show(formatbarhandlebox);

	formatbar =
	    gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
	gtk_container_add(GTK_CONTAINER(formatbarhandlebox), formatbar);
	gtk_widget_show(formatbar);
	setup_buttons();
	init_toggle();



	label1 = gtk_label_new("Label 1");
	gtk_misc_set_alignment(GTK_MISC(label1), 0, 0);
	gtk_box_pack_start(GTK_BOX(topbox), label1, FALSE, FALSE, 2);
	gtk_widget_show(label1);

	text1 = gtk_entry_new();
	GTK_WIDGET_UNSET_FLAGS(text1, GTK_CAN_DEFAULT);
	gtk_signal_connect_after(GTK_OBJECT(text1),
				 "key_press_event",
				 (GtkSignalFunc) dialog_key,
				 GTK_OBJECT(text1));
	gtk_box_pack_start(GTK_BOX(topbox), text1, FALSE, FALSE, 2);
	gtk_widget_show(text1);
	gridpane = gtk_vbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(topbox), gridpane, TRUE, TRUE, 2);
	gtk_widget_show(gridpane);

	label2 = gtk_label_new("Label 2");
	gtk_misc_set_alignment(GTK_MISC(label2), 0, 0);
	gtk_box_pack_start(GTK_BOX(topbox), label2, FALSE, FALSE, 2);
	gtk_widget_show(label2);
}

void activate_window(window * w)
{
	char b[256];

	if (w_list)
		GTK_SGRID(w_list->ui->grid)->visible_cursor = FALSE;
	w_list = w;
	strcpy(b, "Siag: ");
	strncat(b, w->buf->name, 200);
	gtk_window_set_title(GTK_WINDOW(topLevel), b);
	GTK_SGRID(w->ui->grid)->visible_cursor = TRUE;
	gtk_widget_grab_focus(w->ui->grid);	/* grab_default? */
}

window *find_window_by_widget(GtkWidget * wdg)
{
	window *w = w_list;
	do {
		if (w->ui->viewport == wdg || w->ui->selectall == wdg ||
		    w->ui->colnum == wdg || w->ui->rownum == wdg ||
		    w->ui->grid == wdg || w->ui->vscroll == wdg ||
		    w->ui->hscroll == wdg)
			return w;
		w = w->next;
	} while (w != w_list);
	return NULL;
}

void free_window(window * w)
{
	window *pw;

	for (pw = w_list; pw->next != w && pw->next != pw; pw = pw->next);
	pw->next = w->next;

	if (w_list == w)
		w_list = w_list->next;
	if (w_list == w)
		w_list = NULL;
	gtk_widget_destroy(w->ui->viewport);
	cfree(w->ui);
	cfree(w);
}

static gchar *ng_colnum_text(GtkSGrid * dummy, gpointer data, gint row,
			     gint col)
{
	static char b[80];
	sprintf(b, "%d", col);
	return b;
}

static gchar *ng_rownum_text(GtkSGrid * dummy, gpointer data, gint row,
			     gint col)
{
	static char b[80];
	sprintf(b, "%d", row);
	return b;
}

static int colnum_click(GtkWidget * w, GdkEventButton * event,
			gpointer user_data)
{
	gint tx, ty, row, col;
	hide_cur(w_list);
	activate_window(find_window_by_widget(w));
	tx = event->x;
	ty = event->y;
	gtk_sgrid_coord_to_rowcol(GTK_SGRID(w), tx, ty, &row, &col);
	set_point_col(w_list, col);
	set_block(1, col, BUFFER_ROWS, col);
	pr_scr_flag = TRUE;
	show_cur(w_list);
	return TRUE;
}

static int rownum_click(GtkWidget * w, GdkEventButton * event,
			gpointer user_data)
{
	gint tx, ty, row, col;
	hide_cur(w_list);
	activate_window(find_window_by_widget(w));
	tx = event->x;
	ty = event->y;
	gtk_sgrid_coord_to_rowcol(GTK_SGRID(w), tx, ty, &row, &col);
	set_point_row(w_list, row);
	set_block(row, 1, row, BUFFER_COLS);
	pr_scr_flag = TRUE;
	show_cur(w_list);
	return TRUE;
}

static int grid_click(GtkWidget * w, GdkEventButton * event,
		      gpointer user_data)
{
	gint tx, ty, row, col;
	hide_cur(w_list);
	activate_window(find_window_by_widget(w));
	tx = event->x;
	ty = event->y;
	gtk_sgrid_coord_to_rowcol(GTK_SGRID(w), tx, ty, &row, &col);
	set_point_row(w_list, row);
	set_point_col(w_list, col);
	pr_scr_flag = TRUE;
	show_cur(w_list);
	return TRUE;
}

/*#define CTRL(c) ((c)&31)
*/
#define CONTROL_MASK 4
/*#define ALT(c) ((c)|0x80)
*/
#define ALT_MASK 8

static int grid_key(GtkWidget * w, GdkEventKey * event, gpointer user_data)
{
	lastc = event->keyval;

	switch (lastc) {
	case GDK_Home:
		lastc = CTRL('a');
		break;
	case GDK_Left:
		lastc = CTRL('b');
		break;
	case GDK_Delete:
		lastc = CTRL('d');
		break;
	case GDK_End:
		lastc = CTRL('e');
		break;
	case GDK_Right:
		lastc = CTRL('f');
		break;
	case GDK_Down:
		lastc = CTRL('n');
		break;
	case GDK_Up:
		lastc = CTRL('p');
		break;
	case GDK_Page_Up:
		lastc = ALT('v');
		break;
	case GDK_Page_Down:
		lastc = CTRL('v');
		break;
	default:
		if (lastc > 255)
			return TRUE;
	}
	if (event->state & GDK_CONTROL_MASK)
		lastc = CTRL(lastc);
	if (event->state & ALT_MASK)
		lastc = ALT(lastc);
	do_cmd(lastc);
	return TRUE;
}
/* These signal handlers for the scrollbars know about the rownum
   and colnum widgets, so that they are scrolled along with the
   grid. Point is also moved to keep it visible. */
static void siag_hadjustment_value_changed(GtkAdjustment * adjustment,
					   gpointer data)
{
	window *w;
	int s;

	g_return_if_fail(adjustment != NULL);
	g_return_if_fail(data != NULL);

	w = (window *) data;
	s = w->sht;
	w->buf->sht[s].top.col = w->buf->sht[s].point_pos.col = adjustment->value;
	draw_cells(w);
}

static void siag_vadjustment_value_changed(GtkAdjustment * adjustment,
					   gpointer data)
{
	window *w;
	int s;

	g_return_if_fail(adjustment != NULL);
	g_return_if_fail(data != NULL);

	w = (window *) data;
	s = w->sht;
	w->buf->sht[s].top.row = w->buf->sht[s].point_pos.row = adjustment->value;
	draw_cells(w);
}

static void siag_set_hadjustment(window * w, GtkAdjustment * adjustment)
{
	g_return_if_fail(w != NULL);
	g_return_if_fail(adjustment != NULL);

	if (w->ui->hadjustment) {
		gtk_signal_disconnect_by_data(GTK_OBJECT
					      (w->ui->hadjustment),
					      (gpointer) w);
		gtk_object_unref(GTK_OBJECT(w->ui->hadjustment));
	}

	w->ui->hadjustment = adjustment;
	gtk_object_ref(GTK_OBJECT(w->ui->hadjustment));

	gtk_signal_connect(GTK_OBJECT(adjustment), "value_changed",
			   (GtkSignalFunc) siag_hadjustment_value_changed,
			   (gpointer) w);
}

static void siag_set_vadjustment(window * w, GtkAdjustment * adjustment)
{
	g_return_if_fail(w != NULL);
	g_return_if_fail(adjustment != NULL);

	if (w->ui->vadjustment) {
		gtk_signal_disconnect_by_data(GTK_OBJECT
					      (w->ui->vadjustment),
					      (gpointer) w);
		gtk_object_unref(GTK_OBJECT(w->ui->vadjustment));
	}

	w->ui->vadjustment = adjustment;
	gtk_object_ref(GTK_OBJECT(w->ui->vadjustment));

	gtk_signal_connect(GTK_OBJECT(adjustment), "value_changed",
			   (GtkSignalFunc) siag_vadjustment_value_changed,
			   (gpointer) w);
}

static int colnum_height()
{
	return w_list->ui->selectall->allocation.height;
}

window *new_window(buffer * b, window * prev)
{
	window *w;
	int h1 = 20, w1 = 60;
	GtkAdjustment *hadjustment, *vadjustment;

	TRACEME((f, "new_window()\n"));

	w = (window *) cmalloc(sizeof(window));

	if (w == NULL)
		return NULL;

	w->ui = (siag_ui *) cmalloc(sizeof(siag_ui));
	if (w->ui == NULL) {
		cfree(w);
		return NULL;
	}

	w->buf = b;
	w->sht = w->bsht = 0;

	if (prev == NULL)
		prev = w;
	else
		w->next = prev->next;
	prev->next = w;

	w->ui->viewport = gtk_table_new(3, 3, FALSE);
	gtk_table_set_row_spacings(GTK_TABLE(w->ui->viewport), 0);
	gtk_box_pack_start(GTK_BOX(gridpane), w->ui->viewport,
			   TRUE, TRUE, 0);
	gtk_widget_show(w->ui->viewport);

	w->ui->selectall = gtk_button_new();
	GTK_WIDGET_UNSET_FLAGS(w->ui->selectall, GTK_CAN_FOCUS);
	gtk_widget_set_usize(w->ui->selectall, 60, 20);
	gtk_table_attach(GTK_TABLE(w->ui->viewport),
			 w->ui->selectall, 0, 1, 0, 1,
			 GTK_FILL, GTK_FILL, 0, 0);
	gtk_signal_connect_object(GTK_OBJECT(w->ui->selectall),
				  "clicked",
				  GTK_SIGNAL_FUNC(execute_callback),
				  (gpointer) "(select-all)");
	gtk_tooltips_set_tip(helptext, w->ui->selectall,
			     "Select everything", NULL);
	gtk_widget_show(w->ui->selectall);

	w->ui->colnum = gtk_sgrid_new(1, BUFFER_COLS);
	GTK_SGRID(w->ui->colnum)->default_height = h1;
	GTK_SGRID(w->ui->colnum)->row_height = colnum_height;
	GTK_SGRID(w->ui->colnum)->col_width = ng_col_width;
	GTK_SGRID(w->ui->colnum)->text = ng_colnum_text;
	GTK_SGRID(w->ui->colnum)->max_row = 1;
	GTK_SGRID(w->ui->colnum)->max_col = BUFFER_COLS;
	GTK_SGRID(w->ui->colnum)->data = b;
#if 1				/* lethal */
	gtk_signal_connect(GTK_OBJECT(w->ui->colnum), "button_press_event",
			   (GtkSignalFunc) colnum_click,
			   GTK_OBJECT(w->ui->colnum));
#endif
	gtk_table_attach(GTK_TABLE(w->ui->viewport),
			 w->ui->colnum, 1, 2, 0, 1,
			 GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
	GTK_SGRID(w->ui->colnum)->bg =
	    &w->ui->colnum->style->bg[GTK_STATE_NORMAL];
	gtk_widget_show(w->ui->colnum);

	w->ui->rownum = gtk_sgrid_new(BUFFER_ROWS, 1);
	GTK_SGRID(w->ui->rownum)->default_width = w1;
	GTK_SGRID(w->ui->rownum)->row_height = ng_row_height;
	GTK_SGRID(w->ui->rownum)->text = ng_rownum_text;
	GTK_SGRID(w->ui->rownum)->max_row = BUFFER_ROWS;
	GTK_SGRID(w->ui->rownum)->max_col = 1;
	GTK_SGRID(w->ui->rownum)->data = b;
	gtk_signal_connect(GTK_OBJECT(w->ui->rownum), "button_press_event",
			   (GtkSignalFunc) rownum_click,
			   GTK_OBJECT(w->ui->rownum));
	gtk_table_attach(GTK_TABLE(w->ui->viewport), w->ui->rownum, 0, 1,
			 1, 2, GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
	GTK_SGRID(w->ui->rownum)->bg =
	    &w->ui->rownum->style->bg[GTK_STATE_NORMAL];
	gtk_widget_show(w->ui->rownum);

	w->ui->grid = gtk_sgrid_new(BUFFER_ROWS, BUFFER_COLS);
	GTK_SGRID(w->ui->grid)->row_height = ng_row_height;
	GTK_SGRID(w->ui->grid)->col_width = ng_col_width;
	GTK_SGRID(w->ui->grid)->format = cell_format;
	GTK_SGRID(w->ui->grid)->text = ng_text;
	GTK_SGRID(w->ui->grid)->max_row = BUFFER_ROWS;
	GTK_SGRID(w->ui->grid)->max_col = BUFFER_COLS;
	GTK_SGRID(w->ui->grid)->data = b;
	GTK_WIDGET_SET_FLAGS(w->ui->grid, GTK_CAN_FOCUS);
	GTK_WIDGET_SET_FLAGS(w->ui->grid, GTK_CAN_DEFAULT);
	gtk_signal_connect(GTK_OBJECT(w->ui->grid), "button_press_event",
			   (GtkSignalFunc) grid_click,
			   GTK_OBJECT(w->ui->grid));
	gtk_signal_connect(GTK_OBJECT(w->ui->grid), "key_press_event",
			   (GtkSignalFunc) grid_key,
			   GTK_OBJECT(w->ui->grid));
	gtk_table_attach(GTK_TABLE(w->ui->viewport), w->ui->grid, 1, 2, 1,
			 2, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL,
			 0, 0);
	gtk_widget_show(w->ui->grid);
	w->ui->hadjustment = w->ui->vadjustment = NULL;
	hadjustment = (GtkAdjustment *) gtk_adjustment_new(1, 1,
							   BUFFER_COLS, 1,
							   1, 5);
	vadjustment =
	    (GtkAdjustment *) gtk_adjustment_new(1, 1, BUFFER_ROWS, 1, 1,
						 5);
	siag_set_hadjustment(w, hadjustment);
	siag_set_vadjustment(w, vadjustment);

	w->ui->vscroll = gtk_vscrollbar_new(w->ui->vadjustment);
	GTK_WIDGET_UNSET_FLAGS(w->ui->vscroll, GTK_CAN_FOCUS);
	gtk_range_set_update_policy(GTK_RANGE(w->ui->vscroll),
				    GTK_UPDATE_CONTINUOUS);
	gtk_table_attach(GTK_TABLE(w->ui->viewport), w->ui->vscroll, 2, 3,
			 0, 2, 0, GTK_EXPAND | GTK_FILL, 0, 0);
	gtk_widget_show(w->ui->vscroll);

	w->ui->hscroll = gtk_hscrollbar_new(w->ui->hadjustment);
	GTK_WIDGET_UNSET_FLAGS(w->ui->hscroll, GTK_CAN_FOCUS);
	gtk_range_set_update_policy(GTK_RANGE(w->ui->hscroll),
				    GTK_UPDATE_CONTINUOUS);
	gtk_table_attach(GTK_TABLE(w->ui->viewport), w->ui->hscroll, 0, 2,
			 2, 3, GTK_EXPAND | GTK_FILL, 0, 0, 0);
	gtk_widget_show(w->ui->hscroll);

	return w;
}

static LISP bogus()
{
	TRACEME((f, "bogus()"));
	return NIL;
}

void init_windows(buffer * b, int *argc, char **argv)
{
	char *p;
	int s;

	init_windows1(argc, argv);
	interp_startup();

	gtk_widget_show(topLevel);
	activate_window(new_window(b, NULL));
	s = w_list->sht;

	p = ret_text(w_list->buf, s,
		     w_list->buf->sht[s].point_pos.row,
		     w_list->buf->sht[s].point_pos.col);
	if (p == NULL)
		p = "";
	draw_input(p);
	draw_status("");
	draw_cells(w_list);

	init_subr_1("tooltip-mode", ltooltip_mode);
	init_subr_1("add-menu", add_menu);
	init_subr_3("add-menu-entry", add_menu_entry);
	init_subr_2("add-submenu", add_submenu);
	init_subr_4("add-submenu-entry", add_submenu_entry);
	init_subr_3("plugin-register", bogus);
	init_subr_1("helptext-mode", bogus);
	init_calc_cmds();
}

void exit_windows()
{
	/* free all buffers */
	while (b_list != NULL)
		free_buffer(b_list);
	while (w_list != NULL)
		free_window(w_list);
}

static void pr_scr()
{
	window *w;
	buffer *b;

	draw_status("");
	b = b_list;
	do {
		if (b->recalc) {
			int i;
			b->recalc = 0;
			for (i = 0; i < recalc; i++)
				calc_matrix(b);
		}
		b = b->next;
	} while (b != b_list);

	w = w_list;
	do {
		draw_cells(w);
		w = w->next;
	} while (w != w_list);
	pr_scr_flag = FALSE;
}				/* pr_scr */


void show_format(void)
{
	static int last_fmt = -1;
	int s = w_list->sht;
	int fmt = ret_format(w_list->buf, s,
			     w_list->buf->sht[s].point_pos.row,
			     w_list->buf->sht[s].point_pos.col);

	last_fmt = fmt;
}

int cursor_visible = FALSE;

void show_cur(window * w)
{
	int s = w->sht;
	char *p;

	gtk_sgrid_set_point(GTK_SGRID(w->ui->grid),
			    w->buf->sht[s].point_pos.row, w->buf->sht[s].point_pos.col);
	if (pr_scr_flag) {
		pr_scr();
	} else {
		/* this may have moved the top, so we must check that */
		w->buf->sht[s].top.row = GTK_SGRID(w->ui->grid)->top_row;
		w->buf->sht[s].top.col = GTK_SGRID(w->ui->grid)->top_col;
	}
	p = ret_text(w->buf, s,
		     w->buf->sht[s].point_pos.row,
		     w->buf->sht[s].point_pos.col);
	if (p == NULL)
		p = "";
	draw_input(p);
	show_format();
	cursor_visible = TRUE;
}				/* show_cur */

